#include <ruby.h>
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

typedef struct _rbCurl  rbCurl;
typedef struct _pFILE   pFILE;

struct _pFILE {
    char   *pt;
    size_t len;
};

struct _rbCurl {
    CURL   *curl;
    VALUE  prms;    /* ref to Hash */

    VALUE  body_str;   /* strings */
    VALUE  header_str; /* strings */
    VALUE  input_str;  /* strings */
    VALUE  progfunc;  /* ref to Block */
    VALUE  passfunc;  /* ref to Block */

    pFILE  fin;                /* pseudo FILE */
    FILE   *file_file;         /* when use FILE */
    FILE   *file_writeheader;  /* when use FILE */
    FILE   *file_infile;       /* when use FILE */
    FILE   *file_stderr;       /* when use FILE */

    struct curl_slist  *list_httpheader;
    struct curl_slist  *list_quote;
    struct curl_slist  *list_postquote;
    struct HttpPost    *list_httppost;
};

#define GetRbCurl(self, pt)  Data_Get_Struct(self, rbCurl, pt)

static VALUE  cCurl;

/***************************************************************/
/***************************************************************/
/* Auto func begin */
static VALUE
set_url(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("url"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_URL, dst);
    return Qnil;
}
static VALUE
get_url(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("url"));
}

static VALUE
set_proxy(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("proxy"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_PROXY, dst);
    return Qnil;
}
static VALUE
get_proxy(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("proxy"));
}

static VALUE
set_proxyport(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    if (TYPE(v) != T_FIXNUM) { rb_raise(rb_eArgError, "must be Fixnum"); }
    rb_hash_aset(s->prms, rb_str_new2("proxyport"), v);
    curl_easy_setopt(s->curl, CURLOPT_PROXYPORT, FIX2LONG(v));
    return Qnil;
}
static VALUE
get_proxyport(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("proxyport"));
}

static VALUE
set_userpwd(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("userpwd"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_USERPWD, dst);
    return Qnil;
}
static VALUE
get_userpwd(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("userpwd"));
}

static VALUE
set_useragent(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("useragent"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_USERAGENT, dst);
    return Qnil;
}
static VALUE
get_useragent(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("useragent"));
}

static VALUE
set_cookie(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("cookie"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_COOKIE, dst);
    return Qnil;
}
static VALUE
get_cookie(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("cookie"));
}

static VALUE
set_postfields(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("postfields"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_POSTFIELDS, dst);
    return Qnil;
}
static VALUE
get_postfields(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("postfields"));
}

static VALUE
set_followlocation(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("followlocation"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_FOLLOWLOCATION, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_FOLLOWLOCATION, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_followlocation(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("followlocation"));
}

static VALUE
set_timeout(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    if (TYPE(v) != T_FIXNUM) { rb_raise(rb_eArgError, "must be Fixnum"); }
    rb_hash_aset(s->prms, rb_str_new2("timeout"), v);
    curl_easy_setopt(s->curl, CURLOPT_TIMEOUT, FIX2LONG(v));
    return Qnil;
}
static VALUE
get_timeout(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("timeout"));
}

static VALUE
set_proxyuserpwd(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("proxyuserpwd"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_PROXYUSERPWD, dst);
    return Qnil;
}
static VALUE
get_proxyuserpwd(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("proxyuserpwd"));
}

static VALUE
set_httpproxytunnel(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("httpproxytunnel"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_HTTPPROXYTUNNEL, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_HTTPPROXYTUNNEL, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_httpproxytunnel(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("httpproxytunnel"));
}

static VALUE
set_failonerror(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("failonerror"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_FAILONERROR, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_FAILONERROR, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_failonerror(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("failonerror"));
}

static VALUE
set_referer(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("referer"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_REFERER, dst);
    return Qnil;
}
static VALUE
get_referer(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("referer"));
}

static VALUE
set_sslcert(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("sslcert"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_SSLCERT, dst);
    return Qnil;
}
static VALUE
get_sslcert(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("sslcert"));
}

static VALUE
set_sslcertpasswd(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("sslcertpasswd"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_SSLCERTPASSWD, dst);
    return Qnil;
}
static VALUE
get_sslcertpasswd(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("sslcertpasswd"));
}

static VALUE
set_sslversion(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    if (TYPE(v) != T_FIXNUM) { rb_raise(rb_eArgError, "must be Fixnum"); }
    rb_hash_aset(s->prms, rb_str_new2("sslversion"), v);
    curl_easy_setopt(s->curl, CURLOPT_SSLVERSION, FIX2LONG(v));
    return Qnil;
}
static VALUE
get_sslversion(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("sslversion"));
}

static VALUE
set_cookiefile(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("cookiefile"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_COOKIEFILE, dst);
    return Qnil;
}
static VALUE
get_cookiefile(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("cookiefile"));
}

static VALUE
set_customrequest(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("customrequest"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_CUSTOMREQUEST, dst);
    return Qnil;
}
static VALUE
get_customrequest(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("customrequest"));
}

static VALUE
set_timecondition(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    if (TYPE(v) != T_FIXNUM) { rb_raise(rb_eArgError, "must be Fixnum"); }
    rb_hash_aset(s->prms, rb_str_new2("timecondition"), v);
    curl_easy_setopt(s->curl, CURLOPT_TIMECONDITION, FIX2LONG(v));
    return Qnil;
}
static VALUE
get_timecondition(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("timecondition"));
}

static VALUE
set_timevalue(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    if (TYPE(v) != T_FIXNUM) { rb_raise(rb_eArgError, "must be Fixnum"); }
    rb_hash_aset(s->prms, rb_str_new2("timevalue"), v);
    curl_easy_setopt(s->curl, CURLOPT_TIMEVALUE, FIX2LONG(v));
    return Qnil;
}
static VALUE
get_timevalue(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("timevalue"));
}

static VALUE
set_maxredirs(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    if (TYPE(v) != T_FIXNUM) { rb_raise(rb_eArgError, "must be Fixnum"); }
    rb_hash_aset(s->prms, rb_str_new2("maxredirs"), v);
    curl_easy_setopt(s->curl, CURLOPT_MAXREDIRS, FIX2LONG(v));
    return Qnil;
}
static VALUE
get_maxredirs(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("maxredirs"));
}

static VALUE
set_netrc(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("netrc"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_NETRC, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_NETRC, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_netrc(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("netrc"));
}

static VALUE
set_ftplistonly(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("ftplistonly"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_FTPLISTONLY, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_FTPLISTONLY, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_ftplistonly(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("ftplistonly"));
}

static VALUE
set_ftpappend(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("ftpappend"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_FTPAPPEND, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_FTPAPPEND, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_ftpappend(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("ftpappend"));
}

static VALUE
set_transfertext(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("transfertext"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_TRANSFERTEXT, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_TRANSFERTEXT, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_transfertext(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("transfertext"));
}

static VALUE
set_range(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("range"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_RANGE, dst);
    return Qnil;
}
static VALUE
get_range(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("range"));
}

static VALUE
set_ftpport(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("ftpport"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_FTPPORT, dst);
    return Qnil;
}
static VALUE
get_ftpport(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("ftpport"));
}

static VALUE
set_low_speed_limit(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    if (TYPE(v) != T_FIXNUM) { rb_raise(rb_eArgError, "must be Fixnum"); }
    rb_hash_aset(s->prms, rb_str_new2("low_speed_limit"), v);
    curl_easy_setopt(s->curl, CURLOPT_LOW_SPEED_LIMIT, FIX2LONG(v));
    return Qnil;
}
static VALUE
get_low_speed_limit(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("low_speed_limit"));
}

static VALUE
set_low_speed_time(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    if (TYPE(v) != T_FIXNUM) { rb_raise(rb_eArgError, "must be Fixnum"); }
    rb_hash_aset(s->prms, rb_str_new2("low_speed_time"), v);
    curl_easy_setopt(s->curl, CURLOPT_LOW_SPEED_TIME, FIX2LONG(v));
    return Qnil;
}
static VALUE
get_low_speed_time(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("low_speed_time"));
}

static VALUE
set_resume_from(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    if (TYPE(v) != T_FIXNUM) { rb_raise(rb_eArgError, "must be Fixnum"); }
    rb_hash_aset(s->prms, rb_str_new2("resume_from"), v);
    curl_easy_setopt(s->curl, CURLOPT_RESUME_FROM, FIX2LONG(v));
    return Qnil;
}
static VALUE
get_resume_from(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("resume_from"));
}

static VALUE
set_crlf(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("crlf"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_CRLF, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_CRLF, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_crlf(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("crlf"));
}

static VALUE
set_interface(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("interface"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_INTERFACE, dst);
    return Qnil;
}
static VALUE
get_interface(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("interface"));
}

static VALUE
set_krb4level(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("krb4level"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_KRB4LEVEL, dst);
    return Qnil;
}
static VALUE
get_krb4level(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("krb4level"));
}

static VALUE
set_ssl_verifypeer(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("ssl_verifypeer"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_SSL_VERIFYPEER, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_SSL_VERIFYPEER, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_ssl_verifypeer(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("ssl_verifypeer"));
}

static VALUE
set_cainfo(VALUE self, VALUE v)
{
    rbCurl *s;
    char   *src, *dst;
    int    len;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("cainfo"), v);  /* keep obj */

    src = rb_str2cstr(v, &len);   /* copy string */
    dst = ALLOC_N(char, len+1);   /* able to free */
    memcpy(dst, src, len);
    dst[len] = '\0';

    curl_easy_setopt(s->curl, CURLOPT_CAINFO, dst);
    return Qnil;
}
static VALUE
get_cainfo(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("cainfo"));
}

static VALUE
set_verbose(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("verbose"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_VERBOSE, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_VERBOSE, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_verbose(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("verbose"));
}

static VALUE
set_httpheader(VALUE self, VALUE v)
{
    rbCurl *s;
    int i;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("httpheader"), v);  /* keep obj */

    curl_slist_free_all(s->list_httpheader);
    for (i = 0; i < RARRAY(v)->len; i++) {
        s->list_httpheader = curl_slist_append(s->list_httpheader, rb_str2cstr(RARRAY(v)->ptr[i], 0));
    }
    return Qnil;
}
static VALUE
get_httpheader(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("httpheader"));
}

static VALUE
set_quote(VALUE self, VALUE v)
{
    rbCurl *s;
    int i;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("quote"), v);  /* keep obj */

    curl_slist_free_all(s->list_quote);
    for (i = 0; i < RARRAY(v)->len; i++) {
        s->list_quote = curl_slist_append(s->list_quote, rb_str2cstr(RARRAY(v)->ptr[i], 0));
    }
    return Qnil;
}
static VALUE
get_quote(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("quote"));
}

static VALUE
set_postquote(VALUE self, VALUE v)
{
    rbCurl *s;
    int i;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("postquote"), v);  /* keep obj */

    curl_slist_free_all(s->list_postquote);
    for (i = 0; i < RARRAY(v)->len; i++) {
        s->list_postquote = curl_slist_append(s->list_postquote, rb_str2cstr(RARRAY(v)->ptr[i], 0));
    }
    return Qnil;
}
static VALUE
get_postquote(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("postquote"));
}

static VALUE
set_header(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("header"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_HEADER, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_HEADER, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_header(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("header"));
}

static VALUE
set_nobody(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("nobody"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_NOBODY, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_NOBODY, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_nobody(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("nobody"));
}

static VALUE
set_put(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("put"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_PUT, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_PUT, FALSE); break; }
    default:	 { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_put(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("put"));
}

static VALUE
set_upload(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("upload"), v);

    switch (v) {
    case Qtrue:  { curl_easy_setopt(s->curl, CURLOPT_UPLOAD, TRUE); break; }
    case Qfalse: { curl_easy_setopt(s->curl, CURLOPT_UPLOAD, FALSE); break; }
    default:     { rb_raise(rb_eArgError, "must be true or false"); }
    }
    return Qnil;
}
static VALUE
get_upload(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("upload"));
}

static VALUE
set_file(VALUE self, VALUE v)
{
    rbCurl *s;
    int i;
    FILE *f;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("file"), v);  /* keep obj */

    f = fopen(rb_str2cstr(v, 0), "w");
    if (!f) {
        rb_raise(rb_eRuntimeError, "can't open file");
    }
    curl_easy_setopt(s->curl, CURLOPT_FILE, f);
    s->file_file = f;
    return Qnil;
}
static VALUE
get_file(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("file"));
}

static VALUE
set_writeheader(VALUE self, VALUE v)
{
    rbCurl *s;
    int i;
    FILE *f;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("writeheader"), v);  /* keep obj */

    f = fopen(rb_str2cstr(v, 0), "w");
    if (!f) {
        rb_raise(rb_eRuntimeError, "can't open file");
    }
    curl_easy_setopt(s->curl, CURLOPT_WRITEHEADER, f);
    s->file_writeheader = f;
    return Qnil;
}
static VALUE
get_writeheader(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("writeheader"));
}

static VALUE
set_infile(VALUE self, VALUE v)
{
    rbCurl *s;
    int i;
    FILE *f;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("infile"), v);  /* keep obj */

    f = fopen(rb_str2cstr(v, 0), "r");
    if (!f) {
        rb_raise(rb_eRuntimeError, "can't open file");
    }
    curl_easy_setopt(s->curl, CURLOPT_INFILE, f);
    s->file_infile = f;
    return Qnil;
}
static VALUE
get_infile(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("infile"));
}

static VALUE
set_stderr(VALUE self, VALUE v)
{
    rbCurl *s;
    int i;
    FILE *f;
    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("stderr"), v);  /* keep obj */

    f = fopen(rb_str2cstr(v, 0), "w");
    if (!f) {
        rb_raise(rb_eRuntimeError, "can't open file");
    }
    curl_easy_setopt(s->curl, CURLOPT_STDERR, f);
    s->file_stderr = f;
    return Qnil;
}
static VALUE
get_stderr(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("stderr"));
}

/* Auto func end */

/************************************************************************/

static VALUE
set_httppost(VALUE self, VALUE v)
{
    rbCurl *s;
    int    i;
    struct HttpPost *last = NULL;

    Data_Get_Struct(self, rbCurl, s);
    rb_hash_aset(s->prms, rb_str_new2("httppost"), v);  /* keep obj */

    if (s->list_httppost) { curl_formfree(s->list_httppost); }
    for (i = 0; i < RARRAY(v)->len; i++) {
        curl_formparse(rb_str2cstr(RARRAY(v)->ptr[i], 0), &s->list_httppost, &last);
    }
    return Qnil;
}
static VALUE
get_httppost(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return rb_hash_aref(s->prms, rb_str_new2("httppost"));
}

/************************************************************************/

static int
progress_func(VALUE *proc, size_t dltotal, size_t dlnow, size_t ultotal, size_t ulnow)
{
    rb_funcall(*proc, rb_intern("call"), 4, INT2FIX(dltotal),INT2FIX(dlnow),INT2FIX(ultotal),INT2FIX(ulnow));
    return 0;
}

static VALUE
set_progressfunction(VALUE self, VALUE v)
{
    rbCurl *s;
    int i;
    Data_Get_Struct(self, rbCurl, s);
    s->progfunc = v;
    curl_easy_setopt(s->curl, CURLOPT_PROGRESSFUNCTION, progress_func);
    curl_easy_setopt(s->curl, CURLOPT_PROGRESSDATA, &(s->progfunc));
    return Qnil;
}
static VALUE
get_progressfunction(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return s->progfunc;
}

/************************************************************************/

static int
passwd_func(VALUE *proc, char *prompt, char *buffer, int buflen)
{
    VALUE passwd;
    char  *str;
    int   len;
    passwd = rb_funcall(*proc, rb_intern("call"), 1, rb_str_new2(prompt));
    str = rb_str2cstr(passwd, &len);
    if (len >= buflen) { return -1; }
    memcpy(buffer, str, len);
    buffer[len] = '\0';
    return 0;
}

static VALUE
set_passwdfunction(VALUE self, VALUE v)
{
    rbCurl *s;
    int i;
    Data_Get_Struct(self, rbCurl, s);
    s->passfunc = v;
    curl_easy_setopt(s->curl, CURLOPT_PASSWDFUNCTION, passwd_func);
    curl_easy_setopt(s->curl, CURLOPT_PASSWDDATA, &(s->passfunc));
    return Qnil;
}
static VALUE
get_passwdfunction(VALUE self, VALUE v)
{
    rbCurl *s;
    Data_Get_Struct(self, rbCurl, s);
    return s->passfunc;
}

/************************************************************************/
/************************************************************************/

static void
mark_rbCurl(rbCurl *s)
{
    rb_gc_mark(s->prms);
    rb_gc_mark(s->body_str);
    rb_gc_mark(s->header_str);
    rb_gc_mark(s->input_str);
    rb_gc_mark(s->progfunc);
    rb_gc_mark(s->passfunc);
}

static void
free_rbCurl(rbCurl *s)
{
    curl_slist_free_all(s->list_httpheader);
    curl_slist_free_all(s->list_quote);
    curl_slist_free_all(s->list_postquote);
    if (s->list_httppost)   { curl_formfree(s->list_httppost); }

    curl_easy_cleanup(s->curl);
    free(s);
}

/************************************************************************/

static size_t
write_func(void *src, size_t size, size_t nmemb, VALUE *rbstr)
{
    long len = size * nmemb;
    rb_str_cat(*rbstr, src, len);
    return len;
}

static size_t 
read_func(void *dst, size_t size, size_t nmemb, pFILE *f)
{
    size_t  max_len = size * nmemb;
    size_t  len;

    if (!f->len) { return 0; }
    len = f->len;
    if (f->len > max_len) { len = max_len; }
    memcpy(dst, f->pt, len);
    f->len -= len;
    f->pt += len;
    return len;
}


static VALUE
new(VALUE klass)
{
    rbCurl  *s;
    VALUE   obj;

    obj = Data_Make_Struct(klass, rbCurl, mark_rbCurl, free_rbCurl, s);
    s->curl = curl_easy_init();
    s->prms = rb_hash_new();

    s->body_str = rb_str_new(0, 0);
    s->header_str = rb_str_new(0, 0);
    s->input_str = rb_str_new(0, 0);
    s->progfunc = Qnil;
    s->passfunc = Qnil;

    s->file_file = NULL;
    s->file_writeheader = NULL;
    s->file_infile = NULL;
    s->file_stderr = NULL;

    s->list_httpheader = NULL;
    s->list_quote = NULL;
    s->list_postquote = NULL;
    s->list_httppost = NULL;

    curl_easy_setopt(s->curl, CURLOPT_MUTE, TRUE);
    return obj;
}

/***************************************************************/

static VALUE
perform(VALUE self)
{
    rbCurl    *s;
    CURLcode  result;
    char      errbuff[CURL_ERROR_SIZE] = "";
    VALUE     v;
    int       len;

    GetRbCurl(self, s);

    v = rb_hash_aref(s->prms, rb_str_new2("postfields"));
    if (v != Qnil) {
	curl_easy_setopt(s->curl, CURLOPT_POSTFIELDSIZE, RSTRING(v)->len);
	curl_easy_setopt(s->curl, CURLOPT_POST, TRUE);      /* use POST */
    } else {
	curl_easy_setopt(s->curl, CURLOPT_POST, FALSE);
    }

    if (s->list_httpheader) { curl_easy_setopt(s->curl, CURLOPT_HTTPHEADER, s->list_httpheader); }
    if (s->list_quote)      { curl_easy_setopt(s->curl, CURLOPT_QUOTE, s->list_quote); }
    if (s->list_postquote)  { curl_easy_setopt(s->curl, CURLOPT_POSTQUOTE, s->list_postquote); }
    if (s->list_httppost)   { curl_easy_setopt(s->curl, CURLOPT_HTTPPOST, s->list_httppost); }

    if (s->progfunc == Qnil) {
	curl_easy_setopt(s->curl, CURLOPT_NOPROGRESS, TRUE);
    } else {
	curl_easy_setopt(s->curl, CURLOPT_NOPROGRESS, FALSE);
    }

    /* No file for body and header */
    if ((!s->file_file) && (!s->file_writeheader)) {
	/* use Ruby String  */
	curl_easy_setopt(s->curl, CURLOPT_FILE, &(s->body_str));
	curl_easy_setopt(s->curl, CURLOPT_WRITEFUNCTION, write_func);
	curl_easy_setopt(s->curl, CURLOPT_WRITEHEADER, &(s->header_str));
	curl_easy_setopt(s->curl, CURLOPT_HEADER, FALSE);
    }

    /* No file for input */
    if (!s->file_infile) {
	/* use Ruby String  */
	s->fin.pt = rb_str2cstr(s->input_str, &len);
	if (len) {  /* input exist */
	    s->fin.len = len;
	    curl_easy_setopt(s->curl, CURLOPT_INFILE, &(s->fin));
	    curl_easy_setopt(s->curl, CURLOPT_READFUNCTION, read_func);
	    curl_easy_setopt(s->curl, CURLOPT_INFILESIZE, len);
	    curl_easy_setopt(s->curl, CURLOPT_UPLOAD, TRUE);
	}
    }

    if (s->file_stderr) {
	curl_easy_setopt(s->curl, CURLOPT_STDERR, s->file_stderr);
    }


    curl_easy_setopt(s->curl, CURLOPT_ERRORBUFFER, errbuff);
    result = curl_easy_perform(s->curl);

    if (s->file_file) { fclose(s->file_file); }
    if (s->file_writeheader) { fclose(s->file_writeheader); }
    if (s->file_infile) { fclose(s->file_infile); }
    if (s->file_stderr) { fclose(s->file_stderr); }

    if (result != CURLE_OK) {
	rb_raise(rb_eRuntimeError, errbuff);
    }

    return self;
}

/************************************************************************/

static VALUE
get_body_str(VALUE self)
{
    rbCurl    *c;
    VALUE     v;
    GetRbCurl(self, c);
    return c->body_str;
}

static VALUE
get_header_str(VALUE self)
{
    rbCurl    *c;
    VALUE     v;
    GetRbCurl(self, c);
    return c->header_str;
}

static VALUE
get_prms(VALUE self)
{
    rbCurl    *s;
    GetRbCurl(self, s);
    return s->prms;
}

static VALUE
set_input_str(VALUE self, VALUE v)
{
    rbCurl *s;
    GetRbCurl(self, s);
    s->input_str = v;
    return Qnil;
}

static VALUE
get_input_str(VALUE self)
{
    rbCurl    *s;
    GetRbCurl(self, s);
    return s->input_str;
}


/***************************************************************/
static VALUE
test(VALUE self, VALUE v)
{
    rbCurl *s;
    GetRbCurl(self, s);

    rb_yield(rb_str_new2("hekomaka"));
    return Qnil;
}

void
Init_curl(void)
{
   cCurl = rb_define_class("Curl", rb_cObject);

   rb_define_singleton_method(cCurl, "new", new, 0);
   rb_define_method(cCurl, "prms",  get_prms, 0);

   rb_define_method(cCurl, "perform", perform, 0);
   rb_define_method(cCurl, "body_str", get_body_str, 0);
   rb_define_method(cCurl, "header_str", get_header_str, 0);
   rb_define_method(cCurl, "input_str=", set_input_str, 1);
   rb_define_method(cCurl, "input_str", get_input_str, 0);

   rb_define_method(cCurl, "httppost=", set_httppost, 1);
   rb_define_method(cCurl, "httppost", get_httppost, 0);

   rb_define_method(cCurl, "progressfunction=", set_progressfunction, 1);
   rb_define_method(cCurl, "progressfunction", get_progressfunction, 0);
   rb_define_method(cCurl, "passwdfunction=", set_passwdfunction, 1);
   rb_define_method(cCurl, "passwdfunction", get_passwdfunction, 0);

   rb_define_method(cCurl, "test", test, 0);

/* Auto define begin */
   rb_define_method(cCurl, "url=", set_url, 1);
   rb_define_method(cCurl, "url", get_url, 0);
   rb_define_method(cCurl, "proxy=", set_proxy, 1);
   rb_define_method(cCurl, "proxy", get_proxy, 0);
   rb_define_method(cCurl, "proxyport=", set_proxyport, 1);
   rb_define_method(cCurl, "proxyport", get_proxyport, 0);
   rb_define_method(cCurl, "userpwd=", set_userpwd, 1);
   rb_define_method(cCurl, "userpwd", get_userpwd, 0);
   rb_define_method(cCurl, "useragent=", set_useragent, 1);
   rb_define_method(cCurl, "useragent", get_useragent, 0);
   rb_define_method(cCurl, "cookie=", set_cookie, 1);
   rb_define_method(cCurl, "cookie", get_cookie, 0);
   rb_define_method(cCurl, "postfields=", set_postfields, 1);
   rb_define_method(cCurl, "postfields", get_postfields, 0);
   rb_define_method(cCurl, "followlocation=", set_followlocation, 1);
   rb_define_method(cCurl, "followlocation", get_followlocation, 0);
   rb_define_method(cCurl, "timeout=", set_timeout, 1);
   rb_define_method(cCurl, "timeout", get_timeout, 0);
   rb_define_method(cCurl, "proxyuserpwd=", set_proxyuserpwd, 1);
   rb_define_method(cCurl, "proxyuserpwd", get_proxyuserpwd, 0);
   rb_define_method(cCurl, "httpproxytunnel=", set_httpproxytunnel, 1);
   rb_define_method(cCurl, "httpproxytunnel", get_httpproxytunnel, 0);
   rb_define_method(cCurl, "failonerror=", set_failonerror, 1);
   rb_define_method(cCurl, "failonerror", get_failonerror, 0);
   rb_define_method(cCurl, "referer=", set_referer, 1);
   rb_define_method(cCurl, "referer", get_referer, 0);
   rb_define_method(cCurl, "sslcert=", set_sslcert, 1);
   rb_define_method(cCurl, "sslcert", get_sslcert, 0);
   rb_define_method(cCurl, "sslcertpasswd=", set_sslcertpasswd, 1);
   rb_define_method(cCurl, "sslcertpasswd", get_sslcertpasswd, 0);
   rb_define_method(cCurl, "sslversion=", set_sslversion, 1);
   rb_define_method(cCurl, "sslversion", get_sslversion, 0);
   rb_define_method(cCurl, "cookiefile=", set_cookiefile, 1);
   rb_define_method(cCurl, "cookiefile", get_cookiefile, 0);
   rb_define_method(cCurl, "customrequest=", set_customrequest, 1);
   rb_define_method(cCurl, "customrequest", get_customrequest, 0);
   rb_define_method(cCurl, "timecondition=", set_timecondition, 1);
   rb_define_method(cCurl, "timecondition", get_timecondition, 0);
   rb_define_method(cCurl, "timevalue=", set_timevalue, 1);
   rb_define_method(cCurl, "timevalue", get_timevalue, 0);
   rb_define_method(cCurl, "maxredirs=", set_maxredirs, 1);
   rb_define_method(cCurl, "maxredirs", get_maxredirs, 0);
   rb_define_method(cCurl, "netrc=", set_netrc, 1);
   rb_define_method(cCurl, "netrc", get_netrc, 0);
   rb_define_method(cCurl, "ftplistonly=", set_ftplistonly, 1);
   rb_define_method(cCurl, "ftplistonly", get_ftplistonly, 0);
   rb_define_method(cCurl, "ftpappend=", set_ftpappend, 1);
   rb_define_method(cCurl, "ftpappend", get_ftpappend, 0);
   rb_define_method(cCurl, "transfertext=", set_transfertext, 1);
   rb_define_method(cCurl, "transfertext", get_transfertext, 0);
   rb_define_method(cCurl, "range=", set_range, 1);
   rb_define_method(cCurl, "range", get_range, 0);
   rb_define_method(cCurl, "ftpport=", set_ftpport, 1);
   rb_define_method(cCurl, "ftpport", get_ftpport, 0);
   rb_define_method(cCurl, "low_speed_limit=", set_low_speed_limit, 1);
   rb_define_method(cCurl, "low_speed_limit", get_low_speed_limit, 0);
   rb_define_method(cCurl, "low_speed_time=", set_low_speed_time, 1);
   rb_define_method(cCurl, "low_speed_time", get_low_speed_time, 0);
   rb_define_method(cCurl, "resume_from=", set_resume_from, 1);
   rb_define_method(cCurl, "resume_from", get_resume_from, 0);
   rb_define_method(cCurl, "crlf=", set_crlf, 1);
   rb_define_method(cCurl, "crlf", get_crlf, 0);
   rb_define_method(cCurl, "interface=", set_interface, 1);
   rb_define_method(cCurl, "interface", get_interface, 0);
   rb_define_method(cCurl, "krb4level=", set_krb4level, 1);
   rb_define_method(cCurl, "krb4level", get_krb4level, 0);
   rb_define_method(cCurl, "ssl_verifypeer=", set_ssl_verifypeer, 1);
   rb_define_method(cCurl, "ssl_verifypeer", get_ssl_verifypeer, 0);
   rb_define_method(cCurl, "cainfo=", set_cainfo, 1);
   rb_define_method(cCurl, "cainfo", get_cainfo, 0);
   rb_define_method(cCurl, "verbose=", set_verbose, 1);
   rb_define_method(cCurl, "verbose", get_verbose, 0);
   rb_define_method(cCurl, "httpheader=", set_httpheader, 1);
   rb_define_method(cCurl, "httpheader", get_httpheader, 0);
   rb_define_method(cCurl, "quote=", set_quote, 1);
   rb_define_method(cCurl, "quote", get_quote, 0);
   rb_define_method(cCurl, "postquote=", set_postquote, 1);
   rb_define_method(cCurl, "postquote", get_postquote, 0);
   rb_define_method(cCurl, "header=", set_header, 1);
   rb_define_method(cCurl, "header", get_header, 0);
   rb_define_method(cCurl, "nobody=", set_nobody, 1);
   rb_define_method(cCurl, "nobody", get_nobody, 0);
   rb_define_method(cCurl, "put=", set_put, 1);
   rb_define_method(cCurl, "put", get_put, 0);
   rb_define_method(cCurl, "upload=", set_upload, 1);
   rb_define_method(cCurl, "upload", get_upload, 0);
   rb_define_method(cCurl, "file=", set_file, 1);
   rb_define_method(cCurl, "file", get_file, 0);
   rb_define_method(cCurl, "writeheader=", set_writeheader, 1);
   rb_define_method(cCurl, "writeheader", get_writeheader, 0);
   rb_define_method(cCurl, "infile=", set_infile, 1);
   rb_define_method(cCurl, "infile", get_infile, 0);
   rb_define_method(cCurl, "stderr=", set_stderr, 1);
   rb_define_method(cCurl, "stderr", get_stderr, 0);
/* Auto define end */

   rb_define_const(cCurl, "TIMECOND_NONE", INT2FIX(TIMECOND_NONE));
   rb_define_const(cCurl, "TIMECOND_IFMODSINCE", INT2FIX(TIMECOND_IFMODSINCE));
   rb_define_const(cCurl, "TIMECOND_IFUNMODSINCE", INT2FIX(TIMECOND_IFUNMODSINCE));
   rb_define_const(cCurl, "TIMECOND_LASTMOD", INT2FIX(TIMECOND_LASTMOD));

}
